'''Assetto Corsa Websocket Server Class'''

import asyncio
import configparser
from dataclasses import dataclass
import hashlib
import logging
import os
import sys
from typing import Dict
import websockets

from ac_websocket_server.constants import HOST, PORT
from ac_websocket_server.error import AssettoCorsaWebsocketsServerError
from ac_websocket_server.game import Server
from ac_websocket_server.handlers import handler


class AssettoCorsaWebsocketsServer:
    '''Represents an Assetto Corsa WebSocket Server.

    Allows control of an Assetto Corsa server with a websockets interface.'''

    def __init__(self,
                 server_directory: str = None,
                 host: str = HOST,
                 port: int = PORT
                 ) -> None:

        self.__logger = logging.getLogger('ac-ws.ws-server')

        self.host = host
        self.port = port

        if not server_directory:
            self.server_directory = os.getcwd()
        else:
            self.server_directory = server_directory

        self.server: Server = None

        self.send_queue = asyncio.Queue()

        self.stop_server: asyncio.Future = None

    async def consumer(self, message):
        if b'start_server' in message:
            self.__logger.info('Received request to start game server')
            try:
                self.game = Server(server_directory=self.server_directory)
                self.game.subscribe(self)
                await self.game.start()
                self.__logger.info('Game server started')
            except AssettoCorsaWebsocketsServerError as e:
                await self.send_queue.put(str(e))
            return
        if b'stop_server' in message:
            self.__logger.info('Received request to stop game server')
            try:
                await self.game.stop()
                self.game.unsubscribe(self)
                self.__logger.info('Game server stopped')
            except AssettoCorsaWebsocketsServerError as e:
                await self.send_queue.put(str(e))
            return

        response_message = f'Received unrecognised message: {message}'
        self.__logger.debug(response_message)
        await self.send_queue.put(response_message)

    async def handler(self, websocket):

        await websocket.send(
            f'Welcome to the Assetto Corsa WebSocker server running at {self.host}:{self.port}')

        await handler(websocket, self.consumer, self.producer)

    async def notify(self, notifier):
        '''Receive a notification of a new message from notifier.
        Pull the data off the notifier queue and process.'''

        message = await notifier.get()
        await self.send_queue.put(message)

    async def producer(self):
        data = await self.send_queue.get()
        return data

    async def start(self):
        '''Start the websocket server'''

        try:

            self.__logger.info(f'Starting websocket server')

            self.stop_server = asyncio.Future()

            async with websockets.serve(self.handler, self.host, self.port):
                await self.stop_server

            self.__logger.info(f'Stopping websocket server')

        except KeyboardInterrupt:
            self.__logger.info(f'Interupting the server')

    async def stop(self):
        '''Stop the websocket server'''

        self.stop_server.set_result()



'''Assetto Corsa Game Server Class'''

import asyncio
from asyncore import read
import configparser
from dataclasses import dataclass, field
from datetime import datetime
from distutils.log import error
from genericpath import isdir
import hashlib
import logging
from mimetypes import init
import os
from posixpath import dirname
import sys
from typing import List
import websockets

from ac_websocket_server.error import AssettoCorsaWebsocketsServerError
from .watcher import Watcher


@dataclass
class Ports:
    '''Represents ports used by AC game server.'''
    http: int
    tcp: int
    udp: int


@dataclass
class Session:
    '''Represents an individual session in the AC game server'''

    type: str
    laps: int
    time: int


@dataclass
class Server:
    '''Represents an Assetto Corsa Server.'''

    server_directory: str

    name: str = field(init=False)
    ports: Ports = field(init=False)
    sessions: List[Session] = field(init=False)

    def __post_init__(self):

        self.__logger = logging.getLogger('ac-ws.game-server')

        self.__cfg: configparser = None

        self.__queue: asyncio.Queue = asyncio.Queue()

        if os.path.exists(f'{self.server_directory}/acServer.py'):
            self.__cwd = None
            self.__exe = f'{self.server_directory}/acServer.py'
            self.__hash = None
        else:
            self.__cwd = self.server_directory
            if sys.platform == 'linux':
                self.__exe = f'{self.server_directory}/acServer'
                self.__hash = 'f781ddfe02e68adfa170b28d0eccbbdc'
            else:
                self.__exe = f'{self.server_directory}/acServer.exe'
                self.__hash = '357e1f1fd8451eac2d567e154f5ee537'

        if os.path.exists(f'{self.server_directory}/cfg/server_cfg.ini'):

            self.__cfg = configparser.ConfigParser()
            self.__cfg.read(f'{self.server_directory}/cfg/server_cfg.ini')

            self.name = self.__cfg['SERVER']['NAME']

            self.ports = Ports(http=self.__cfg['SERVER']['HTTP_PORT'],
                               tcp=self.__cfg['SERVER']['TCP_PORT'],
                               udp=self.__cfg['SERVER']['UDP_PORT'])

            self.sessions = []

        else:
            self.__logger.error(
                f'Missing server_cfg.ini file in {self.server_directory}')

        if not os.path.exists(f'{self.server_directory}/cfg/entry_list.ini'):
            self.__logger.error(
                f'Missing entry_list.ini file in {self.server_directory}')

        self.__logfile_stdout: str
        self.__logfile_stderr: str

        self.__observers = []

        self.__process: asyncio.subprocess.Process

    async def get(self):
        '''Fetch an item from the queue.  Returns None if the queue is empty.'''

        try:
            response = self.__queue.get_nowait()
        except asyncio.QueueEmpty:
            response = None

        return response

    async def notify(self, notifier):
        '''Receive a notification of a new message from log watcher.'''

        message = await notifier.get()
        await self.put(message)

    async def put(self, item):
        '''Put an item on the queu and notify all observers.'''

        await self.__queue.put(item)
        for obs in self.__observers:
            await obs.notify(self)

    async def start(self):
        '''Start the game server.'''

        if self.__hash:
            try:
                with open(self.__exe, 'rb') as file_to_check:
                    data = file_to_check.read()
                    if self.__hash != hashlib.md5(data).hexdigest():
                        raise AssettoCorsaWebsocketsServerError(
                            f'{self.__exe} checksum mismatch')
            except FileNotFoundError:
                raise AssettoCorsaWebsocketsServerError(
                    f'{self.__exe} missing')

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

        self.__logger.info(f'Starting game server')

        os.makedirs(f'{self.server_directory}/logs/session', exist_ok=True)
        os.makedirs(f'{self.server_directory}/logs/error', exist_ok=True)

        self.__logfile_stdout = f'{self.server_directory}/logs/session/output{timestamp}.log'
        self.__logfile_stderr = f'{self.server_directory}/logs/error/error{timestamp}.log'

        session_file = open(self.__logfile_stdout, 'w')
        error_file = open(self.__logfile_stderr, 'w')

        try:
            self.__process = await asyncio.create_subprocess_exec(
                self.__exe, cwd=self.__cwd, stdout=session_file, stderr=error_file)

            self.__logger.info(f'Process pid is: {self.__process.pid}')
            await self.put(f'Assetto Corsa server started')
            await self.put(str(self))
        except PermissionError as e:
            self.__logger.error(f'Process did not start: {e}')
            await self.put(f'Assetto Corsa server did not start')
            raise AssettoCorsaWebsocketsServerError(e)

        self.__watcher_stdout = Watcher(self.__logfile_stdout)
        self.__watcher_stdout.subscribe(self)
        await self.__watcher_stdout.start()

    async def stop(self):
        '''Stop the game server'''

        self.__logger.info(f'Stopping game server')
        await self.put(f'Assetto Corsa server is stopping')

        self.__process.terminate()

        status_code = await asyncio.wait_for(self.__process.wait(), None)
        self.__logger.info(f'Game server exited with {status_code}')

        await self.__watcher_stdout.stop()
        self.__watcher_stdout.unsubscribe(self)

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.
        Observer object must include an async notify(self, observable, *args, **kwargs) method.'''
        self.__observers.append(observer)

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        try:
            self.__observers.remove(observer)
        except ValueError as error:
            self.logger.debug(
                "Unsubscribe failed with value error: %s for %s", error, observer)



'''Assetto Corsa Log Watcher'''

from aiowatcher import AIOWatcher
import asyncio
import os
import re
import sys

from .observer import Notifier


class Watcher(Notifier):
    '''Represents a watcher for AC logfiles.   
    Parses log files and sends messages to send_queue.'''

    def __init__(self, filename: str) -> None:
        '''Create Watcher instance for filename.'''

        self.__stop = asyncio.Event()

        self.directory = os.path.dirname(filename)
        self.filename = os.path.basename(filename)

        self.__session_name: str = None
        self.__session_laps: int = None
        self.__session_time: int = None

        self.__driver_host: str = None
        self.__driver_port: int = None
        self.__driver_name: str = None
        self.__driver_guid: str = None
        self.__driver_car: str = None

        super().__init__()

    async def parse_lines(self, file, lines):

        for line in lines:

            '''
            DynamicTrack: first session, resetting grip
            Weather update. Ambient: 17.255133 Road: 23.6746 Graphics: 3_clear
            Wind update. Speed: 0 Direction: 0
            SENDING session name : Practice
            SENDING session index : 0
            SENDING session type : 1
            SENDING session time : 10
            SENDING session laps : 0
            SENDING http://93.57.10.21/lobby.ashx/ping?session=1&timeleft=600&port=9600&clients=0&track=magione&pickup=1
            OK
            '''
            m = re.compile(
                r'^SENDING session name : (.*)').match(line)
            if m:
                self.__session_name = m.group(1)

            m = re.compile(
                r'^SENDING session time : (.*)').match(line)
            if m:
                self.__session_time = int(m.group(1))

            m = re.compile(
                r'^SENDING session laps : (.*)').match(line)
            if m:
                self.__session_laps = int(m.group(1))
                await self.put(f'Starting {self.__session_name} session ' +
                               f'with {str(self.__session_laps)} laps and {str(self.__session_time)} minutes')

            '''
            NEW PICKUP CONNECTION from  192.168.1.1:50834
            VERSION 202
            Mark Hannon
            REQUESTED CAR: bmw_m3_e30*
            ENTRY LIST OPEN MODE 1
            Looking for available slot by name for GUID 76561198005431407 bmw_m3_e30
            Looking for available slot
            Slot found at index 0
            ResetCarResults, index: 0
            DRIVER ACCEPTED FOR CAR 0
            DRIVER ACCEPTED FOR CAR Mark Hannon
            Tyre blankets: false
            Sending SESSION ID : 0
            Sending type 1
            Sending type 2
            Sending type 3
            Sending ID : 0
            Sending type 1
            FIND AND SEND PLAYER'S POSITION
            ELAPSED=70807
            Sending 3 checksum requests
            SERVER TIME: 73273
            DRIVER: Mark Hannon []
            OK
            '''

            m = re.compile(
                r'^NEW PICKUP CONNECTION from  (.*):(\d*)').match(line)
            if m:
                self.__driver_host = m.group(1)
                self.__driver_port = int(m.group(2))

            m = re.compile(
                r'^Looking for available slot by name for GUID (\d*) (.*)').match(line)
            if m:
                self.__driver_guid = m.group(1)
                self.__driver_car = m.group(2)

            m = re.compile(
                r'^DRIVER: (.*) \[.*$').match(line)
            if m:
                self.__driver_name = m.group(1)
                await self.put(f'{self.__driver_name} [{self.__driver_guid}] is joining in a ' +
                               f'{self.__driver_car} from {self.__driver_host}:{self.__driver_port}')

            '''
            Clean exit, driver disconnected:  Mark Hannon []
            '''

            m = re.compile(
                r'^Clean exit, driver disconnected:  (.*) \[.*$').match(line)
            if m:
                await self.put(f'{m.group(1)} left')

            '''
            ERROR,SERVER NOT REGISTERED WITH LOBBY - PLEASE RESTART
            ERROR - RESTART YOUR SERVER TO REGISTER WITH THE LOBBY
            '''

            m = re.compile(
                r'^ERROR.*$').match(line)
            if m:
                await self.put(f'{line}')

    async def start(self):
        '''Start monitoring logfile'''

        self.__stop.clear()

        while not self.__stop.is_set():
            aiow = AIOWatcher(self.directory,
                              self.parse_lines,
                              filename=self.filename)
            await aiow.init()
            await aiow.loop()

    async def stop(self):
        '''Stop monitoring logfile'''

        self.__stop.set()


async def callback(filename, lines):
    for line in lines:
        print(line[:-1])


async def main():
    lw = AIOWatcher('var', callback, extensions=['txt'])
    await lw.tail(3)
